#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include "basic.h"

static int verbose = 0 ;

#ifdef X86
inline static void cache_flush(void *p) {
asm __volatile__ (
"clflush 0(%[p])\n\t"
: : [p] "r" (p) : "memory");
}
#else
inline static void cache_flush(void *p) {
asm __volatile__ (
"dcbf 0,%[p]\n\t"
: : [p] "r" (p) : "memory");
}
#endif

/* Barrier */

typedef struct {
  unsigned int nprocs ;
  volatile unsigned int count ;
  volatile int sense ;
} barrier_t ; 


barrier_t *alloc_my_barrier(unsigned int nprocs) {
  barrier_t *p = malloc_check(sizeof(*p)) ;
  p->nprocs = p->count = nprocs ;
  p->sense = 0 ;
  return p ;
}

void free_my_barrier(barrier_t *p) {
  free(p) ;
}

/* Coded in C */
int wait_my_barrier(barrier_t *p) {
  int r = 0 ;
  int sense = p->sense ;
  int rem = DECR(&p->count) ;
  SYNC() ;
  if (rem == 0) {
    r = 1 ;
    p->count = p->nprocs ;
    SYNC() ;
    p->sense = !sense ;
  } else {
    unsigned int c = 0 ;
    while (p->sense == sense) { if (++c == 8000000U) { c = 0 ; printf("+") ; fflush(stdout) ; } }
  }
  SYNC() ;
  return r ;
}

#define N 3
#define LINE 1024
#define SZ (LINE/sizeof(int))

typedef struct {
  int r0,r1,r2 ;
} log_t ;

static void pp_log(FILE *chan,log_t *p) {
  fprintf(chan,"r0=%i, r1=%i, r2=%i\n",p->r0,p->r1,p->r2) ;
}

typedef uint64_t tb_t ;
#define PTB PRIu64

#ifdef X86
inline static tb_t read_timebase(void) {
  unsigned int a,d;
  asm __volatile__ ("rdtsc" : "=a" (a), "=d" (d));
  return ((tb_t)a) | (((tb_t)d)<<32);
}
#else
inline static tb_t read_timebase(void) {
  tb_t r;
  asm __volatile__ ("mftb %[r1]" :[r1] "=r" (r) : : "memory");
  return r;
}
#endif

/* Instance context */
typedef struct {
  void *mem ;
  int *cache ;
  log_t *out ;
  tb_t next_tb;
  barrier_t *b;
} ctx_t ;

ctx_t *alloc_ctx(int nlines,int log_sz) {
  ctx_t *r = malloc_check(sizeof(*r)) ;
  r->mem = malloc_check((nlines+1)*LINE) ;
  intptr_t x = (intptr_t)r->mem ;
  x += LINE-1 ;
  x /= LINE ;
  x *= LINE ;
  r->cache = (int *)x ;
  r->out = malloc_check(log_sz*sizeof(*r->out)) ;
  return r ;
}

void free_ctx(ctx_t *p) {
  free(p->mem) ;
  free(p->out) ;
  free(p) ;
}

typedef struct {
  int xp,yp,zp ;
  int d1,d2 ;
} param_t ;


typedef struct {
  int id ;
  int role ;
  barrier_t *b ;
  int size ;
  volatile int *cache ;
  log_t *out ;
  tb_t *next_tb;
  param_t *param ;
} arg_t ;

arg_t *alloc_arg(int size) {
  arg_t *r = malloc_check(sizeof(*r));
  r->size = size ;
  return r ;
}

void free_arg(arg_t*p) { free(p); }

inline static void
dorun(arg_t *a,log_t *out,param_t *p) {
  const int delay0 = 2048 ;
  const int step = 512 ;
  int _delay = 2*delay0 ;

  wait_my_barrier(a->b) ;
  volatile int *x = a->cache + p->xp * SZ ;
  volatile int *y = a->cache + 1 + p->yp * SZ ;
  volatile int *z = a->cache + 2 + p->zp * SZ ;
  cache_flush((void *)x) ;
  cache_flush((void *)y) ;
  cache_flush((void *)z) ; 
  switch (a->role) {
  case 0:
    *x = *y = *z = 0 ;
    break ;
  case 1:
    _delay = delay0 + p->d1*step ;
    break ;
  case 2:
    _delay = delay0 + p->d2*step ;
    break ;
  }
  if (a->role == 0) *a->next_tb = read_timebase() ;
  wait_my_barrier(a->b) ;
  tb_t _tb0 = *a->next_tb;
  int64_t _delta;
  do {
    _delta =  read_timebase() - _tb0;
  } while (_delta < _delay);    
  switch (a->role) {
  case 0:
    *x = 1 ;
    int r0 = *y ;
    out->r0 = r0 ;
    break ;
  case 1:
    *y = 1 ;
    int r1 = *z ;
    out->r1 = r1 ;
    break;
  case 2:
    *z = 1 ;
    int r2 = *x ;
    out->r2 = r2 ;
    break;
  }
  wait_my_barrier(a->b) ;
  if (verbose && a->role == 0) {
    if ((out->r0 == 0 && out->r1 == 0 && out->r2 == 0)) {
      fprintf(stderr,"d1=%i, d2=%i, xp=%i, yp=%i, zp=%i\n",
              p->d1,p->d2,p->xp,p->yp,p->zp) ;
      fflush(stderr) ;
    }
  }
}

void *thread(void *_a) {
  arg_t *a = (arg_t *)_a ;
  param_t *ap = a->param ;
  const int delay0 = 2048 ;
  const int step = 512 ;
  const int dmax = (2*delay0)/step ;

  if (ap == NULL) {
    int k = 0 ;

    for ( ;; ) {
      for (int xp = 0 ; xp < 3 ; xp++)
        for (int yp = 0 ; yp < 3 ; yp++)
          for (int zp = 0 ; zp < 3 ; zp++)
            for (int d1 = 0 ; d1 < dmax ; d1++ )
              for (int d2 = 0 ; d2 < dmax ; d2++) {
                log_t *out = &a->out[k] ;
                param_t p = { xp, yp, zp, d1, d2 } ;
                dorun(a,out,&p) ;
                k++ ;
                if (k >= a->size) goto out;
              }
    }
  } else {
    for (int k = 0 ; k < a->size ; k++) {
      dorun(a,&a->out[k],ap) ;
    }
  }
  out:
  return NULL ;
}

void run(int size, int ninst,param_t *ap) {
  int count[2][2][2] ;
  for (int k1 = 0 ; k1 < 2 ; k1++)
    for (int k2 = 0 ; k2 < 2 ; k2++)
      for (int k3 = 0 ; k3 < 2 ; k3++)
        count[k1][k2][k3] = 0 ;
  int nth = N * ninst ;
  pthread_t th[nth] ;
  ctx_t *ctx[ninst] ;
  arg_t *arg[nth] ;
  int id = 0 ;
  for (int i = 0 ; i < ninst ; i++) {
    ctx[i] = alloc_ctx(N,size) ;
    ctx_t *c = ctx[i] ;
    c->b = alloc_my_barrier(N) ;
    for (int role = 0 ; role < N ; role++) {
      arg[id] = alloc_arg(size) ;
      arg_t *a = arg[id] ;
      a->id = id ; a->role = role ;
      a->b = c->b ; a->cache = c->cache ; a->out = c->out ;
      a->next_tb = &c->next_tb ;
      a->param = ap ;
      id++ ;
    }
  }

  for (int nrun = 1; ; nrun++) {
    for (int n = 0 ; n < nth ; n++)
      create_thread(&th[n],thread,arg[n]) ;

    for (int n = 0 ; n < nth ; n++)
      (void)join_thread(&th[n]) ;

    if (verbose > 1) fprintf(stderr,"+++++++++++++++\n") ;
    for (int k = 0 ; k < size ; k++) {
      for (int i = 0 ; i < ninst ; i++) {
        log_t *p = &ctx[i]->out[k] ;
        if (verbose > 1) pp_log(stderr,p) ;
        count[p->r0][p->r1][p->r2]++ ;
      }
    }
    if (verbose > 1) fflush(stderr) ;

    printf("** %i **\n",nrun) ;
    int nz = 0 ;
    for (int k1 = 0 ; k1 < 2 ; k1++)
      for (int k2 = 0 ; k2 < 2 ; k2++)
        for (int k3 = 0 ; k3 < 2 ; k3++) {
          int c = count[k1][k2][k3] ;
          if (c > 0) nz++ ;
          printf("%-5i> r0=%i r1=%i, r2=%i\n",c,k1,k2,k3) ;
        }
    fflush(stdout) ;
    if (nz == 8) break ;
  }
  for (int i = 0 ; i < ninst ; i++) {
    free_my_barrier(ctx[i]->b) ;
    free_ctx(ctx[i]) ;
  }
  for (int n = 0 ; n < nth ; n++)
    free_arg(arg[n]) ;
}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int size = 16 ;
  
  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {
    size = atoi(argv[0]) ;
  }
  param_t p = { 0, 1, 2, 4, 4, } ;
  run(size,5,NULL) ;
  return 0 ;
}
